import os
import sys
import validators

"""
Utility functions used for the application. Helps to abstract some functions from the other files.
"""

__VALID_EXTENSIONS = ['jpeg', 'jpg', 'png', 'tiff', 'raw', 'webp']

__parent_dir: str = os.path.join(os.path.expanduser('~'), 'Desktop')
__directory: str = 'Ovative Group Caption Generator'
__filename: str = 'outputs.csv'

# The generated directory path used to store outputs (e.g., C:/Users/user_name/Desktop/Ovative Group Caption Generator)
__GEN_PATH: str = os.path.join(__parent_dir, __directory)


def clear():
    os.system('cls' if os.name == 'nt' else 'clear')


def quitting(user_input: str) -> bool:
    return user_input.lower() in ['q', 'quit']


def end() -> None:
    clear()
    print('Ending program.')
    sys.exit()


def break_line():
    """
    A break line that's used to help separate the outputs.
    :return:
    """
    print('\n-----------------------------------------------------------------------')


def is_dir(directory: str) -> bool:
    """
    Checks if the given value is an existing, local directory.

    :param directory: possible directory path
    :return: True if value is an existing directory, False otherwise
    """
    return os.path.exists(directory)


def is_url(image_url: str) -> bool:
    """
    Checks if the given value is a valid URL.

    :param image_url: url of the given image.
    :return:
    """
    return validators.url(image_url)


def clean_url(user_input: str) -> str:
    """
    If the given URL contains 'www,' add 'https://' to the beginning of the string so the model recognizes it.
    :return: None
    """
    if user_input[0:3] == 'www':
        return 'https://' + user_input

    # must specify returning the original input to prevent returning None
    return user_input


def create_dir() -> None:
    """
    Creates the directory the inputs will be stored in and return that directory string.
    :return: string representing the created folder directory
    """

    if not os.path.exists(__GEN_PATH):
        os.makedirs(__GEN_PATH)  # make directory if it doesn't exist already
        print(f'Created a new directory, "{__GEN_PATH}", to store generated captions.')
    else:
        print(f'Generated outputs will be stored in the "{__GEN_PATH}" directory.')


def output_file_path() -> str:
    return os.path.join(__GEN_PATH, __filename)


def validate_and_add_url(img_source: str, image_sources: list[str]) -> None:
    """
    Check if the user's input contains a valid file extension via :func:`is_valid_extension`
    and extend the images into the list being passed in.

    :param image_sources: a list of sources that will be added to if the img_source is valid
    :param img_source: String representing the url of the image.
    :return: none
    """

    extension: str = img_source[::-1].split('.')[0][::-1]

    # check if the file extension is valid
    if is_valid_extension(extension):
        image_sources.append(clean_url(img_source))  # add "https://" if the url starts with "www"; add it to the list
    else:
        extensions = ', '.join(__VALID_EXTENSIONS)
        print(f'Invalid image URL with extension "{extension}" was found in  "{img_source}".',
              'fThis will not be processed. Valid extensions are: "{extensions}".', sep='\n')


def validate_and_add_files(directory: str, image_sources: list[str]) -> None:
    """
    Each file in the given directory is validated to see if the model can use it. If so, its absolute path is written
    to the
    :param directory: Path of the directory with image files to be validated
    :param image_sources: A list that contains all sources for images (urls or filenames)
    :return:
    """

    for file_name in os.listdir(directory):
        # get the file extension by splitting the file name and accessing the extension past the "."
        extension: str = os.path.splitext(file_name)[1].split('.')[1]

        if extension not in __VALID_EXTENSIONS:
            print(f'Invalid image source "{file_name}" in directory: "{directory}" was given.\n'
                  f'This will not be processed.\n')
            continue

        file_path = os.path.join(directory, file_name)
        image_sources.append(file_path)


def is_valid_extension(img_source: str) -> bool:
    """
     Checks if the user's input contains a valid file extension in the given string
     :param img_source: source of the given image with file extension.
     :return: none
     """

    # reverse the string, get the backwards file extension, and then re-reverse the extension back to normal and
    # check if the extension is in list of valid extensions.
    extension: str = img_source[::-1].split('.')[0][::-1]

    valid: bool = extension in __VALID_EXTENSIONS

    return valid


def filter_and_validate(image_sources: list[str]) -> list[str]:
    results: list[str] = []

    for image_source in image_sources:
        if is_url(image_source):
            validate_and_add_url(image_source, results)
        elif os.path.isdir(image_source):
            validate_and_add_files(image_source, results)
        else:
            print(f'{image_source} is not a valid input. This will not be processed.\n')
    return results



import cohere

__co = cohere.Client(
    api_key="52mmmG91mKfuEXARLx4KRzKq448QqGdUiQgDhZNd",
)


def coherence() -> str:
    """
    Returns a coherent output combining the given data.
    :return:
    """
    chat = __co.chat(
        message=f"make a coherent sentence out of list of text {['#WANNASIP?', 'ad', 'SPRITE', 'man with beard sitting next to 6', 'adele', 'purple soda', 'coca cola', 'charlie brown', 'iphone 15', 'iphone14', 'iphone13', 'iphone12']}! Only give the caption without creating an image.",
        model="command"
    )

    return chat.text


if __name__ == "__main__":
    print(coherence())



import easyocr
import pandas as pd
from PIL import Image, ImageDraw
import requests
from io import BytesIO
import numpy as np


def download_image(image_url: str) -> Image:
    """
    Downloads an into a processable format for the OCR. This code was given from the EasyOCR code
    :param image_url: a string representing the URL to the image
    :return: an Image object created from the URL
    """
    response = requests.get(image_url)
    img = Image.open(BytesIO(response.content))
    return img


def draw_boxes(image: Image, bounds, color='yellow', width=2) -> Image:
    """
    Provided code that draws bounding boxes around the text found in the image. This was originally provided by the
    EasyOCR code. It shouldn't be necessary for this project, but is present to ensure the OCR still functions.
    :param image: a Pillow Image object
    :param bounds: a given iterator
    :param color: the color of the boxes to be drawn
    :param width: the width used for drawing the boxes
    :return: the updated Image object with the bounding boxes around the text
    """
    draw = ImageDraw.Draw(image)
    for bound in bounds:
        p0, p1, p2, p3 = bound[0]
        draw.line([*p0, *p1, *p2, *p3, *p0], fill=color, width=width)
    return image


def inference(img_source, lang='en') -> list[str]:
    """
    Reads the text in the image and returns all the text found.
    :param img_source: the source of the image
    :param lang: the language used to detect the text
    :return: the list of text
    """
    if img_source.startswith('http://') or img_source.startswith('https://'):
        im = download_image(img_source)
    else:
        im = Image.open(img_source)

    reader = easyocr.Reader([lang], gpu=False)
    bounds = reader.readtext(np.array(im))
    draw_boxes(im, bounds)
    df = pd.DataFrame(bounds, columns=['Position', 'Text', 'Confidence'])
    filtered_df = df[df['Confidence'] > 0.3]
    return filtered_df['Text'].tolist()



import unittest
"""
This file is used to run the test suite. 

**DO NOT MODIFY THIS FILE.** 
"""

if __name__ == "__main__":
    loader = unittest.TestLoader()
    tests = loader.discover('.')
    testRunner = unittest.TextTestRunner()
    test_results = testRunner.run(tests)

    if test_results.wasSuccessful(): 
        exit(0)
    else:
        exit(1)

# To run the test suite, make sure your terminal is in the root directory of the project
# In the terminal, run 'python -m testing.tests.runner'
# This runs this file as a module of the entire project, allowing imports to function properly



import unittest
from src.utils import *


class TestUtilityFunctions(unittest.TestCase):

    def test_quitting(self):
        self.assertTrue(quitting('q'))
        self.assertTrue(quitting('quit'))
        self.assertFalse(quitting(''))
        self.assertFalse(quitting('quit!'))

    def test_is_dir(self):
        self.assertTrue(is_dir(os.getcwd()))
        self.assertFalse(is_dir('nonexistent_directory'))

    def test_is_url(self):
        self.assertTrue(is_url('https://www.example.com'))
        self.assertFalse(is_url('not_a_url'))

    def test_clean_url(self):
        self.assertEqual(clean_url('www.example.com'), 'https://www.example.com')
        self.assertEqual(clean_url('http://example.com'), None)

    # def test_create_dir(self):
    #     # Mock os.makedirs function
    #     with patch('os.makedirs') as mock_makedirs:
    #         # Call the function under test
    #         directory = create_dir()
    #
    #         # Assert that os.makedirs was called once
    #         mock_makedirs.assert_called_once()
    #
    #         # Assert that the returned directory exists
    #         self.assertTrue(os.path.isdir(directory))

    def test_is_valid_extension(self):
        self.assertTrue(is_valid_extension('jpg'))
        self.assertFalse(is_valid_extension('txt'))

    def test_filter_and_validate(self):
        image_sources = ['https://www.example.com/image.jpg', 'test_images', 'not_a_url']
        filtered_images = filter_and_validate(image_sources)
        self.assertEqual(len(filtered_images), 1)  # Only valid image should be added


if __name__ == '__main__':
    unittest.main()



import csv
import datetime
import os.path

import requests
import torch
from PIL import Image
from transformers import BlipProcessor, BlipForConditionalGeneration

from tqdm import tqdm

from src import utils
from src.models import blip_image as bi, easy_ocr


class ImageProcessor(object):
    """
    This class enforces a singleton structure. This is because only *one* instance of ImageProcessor is needed when the
    application is in use.
    """

    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(ImageProcessor, cls).__new__(ImageProcessor)
        return cls.__instance

    def __source_to_image_object(self, image_sources: list[str]) -> list[tuple[Image, str]]:
        """
        This method converts the URL to an image if it is a valid image link. Otherwise, it converts the given file
        directory.
        :return: Pillow Image type
        """

        # Do not delete this line; used for creating the image
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        results: list[tuple[Image, str]] = []

        for img_source in image_sources:
            if utils.is_url(img_source):
                results.append((Image.open(requests.get(img_source, stream=True).raw).convert('RGB'), img_source))

            if utils.is_dir(img_source):
                results.append(((Image.open(img_source)).convert("RGB"), img_source))

        return results

    def process_input(self, image_sources: list[str]) -> bool:
        """
        This method handles the logic for processing images. It will return a boolean indicating a successful
        processing of the images.
        :param image_sources: A list of image URLs or directory paths to be processed
        :return: True or False
        """
        success = False

        utils.clear()
        print('Checking input...\nThis may take a while if many images were provided.')
        filtered_input: list[str] = utils.filter_and_validate(image_sources)

        processor = BlipProcessor.from_pretrained("Salesforce/blip-image-captioning-large")
        model = BlipForConditionalGeneration.from_pretrained("Salesforce/blip-image-captioning-large")

        # Receive the given input as a Pillow Image object and the given source
        images: list[tuple[Image, str]] | None = self.__source_to_image_object(filtered_input)

        if len(images) == 0:
            print('No provided input was valid. Nothing found to process.')
            return False

        utils.clear()
        print('Processing images...\n')

        utils.create_dir()

        # a list representing the names of each column in the generated .csv file
        fieldnames: list[str] = ['Image Source', 'Conditional Caption', 'Unconditional Caption', 'Text Found in Image',
                                 'Date Processed', 'Time Processed']

        # used to store generated outputs from the models
        # tuples are formed as (generated caption(s), image source)
        rows: list[dict] = []
        values: list[str]

        print()  # to space the captions generated in the terminal

        for img, img_source in tqdm(images, desc='Progress', ascii=False):

            # collect all data/captions from the models *before* adding the information to the rows dict
            captions: tuple[str, str] = bi.caption_image(img, processor, model)

            # get the list of all words found in the image
            ocr_output: list[str] = easy_ocr.inference(img_source)

            # get the current date and time the photos were processed
            date: str = datetime.datetime.now().strftime('%Y-%m-%d')
            current_time: str = datetime.datetime.now().strftime('%H:%M:%S')

            # the values used to store in the rows dict
            values = [img_source, captions[0], captions[1], ocr_output, date, current_time]

            rows.append({k: v for (k, v) in zip(fieldnames, values)})

        success = self.store_outputs(rows)

        return success

    def store_outputs(self, rows: list[dict]) -> bool:
        """
        Store all the given data in their respective columns the .csv file.
        :param rows: a list of dictionaries, where each dict is a row in the .csv file
        :return: True or False for successfully storing the data
        """
        # a boolean representing if the output file exists; using `open` creates the file immediately
        file_exists: bool = os.path.isfile(utils.output_file_path())

        try:
            with open(utils.output_file_path(), 'a') as file:

                # create a writer object to write the given rows in the csv file
                writer = csv.DictWriter(file, fieldnames=rows[0].keys())

                # only write the column headers if the file was created for the first time
                if not file_exists:
                    writer.writeheader()

                for row in rows:
                    writer.writerow(row)

                print('Data stored successfully')

            return True
        except PermissionError:
            utils.clear()
            print(f'WARNING: the file located at "{utils.output_file_path()}" is open and is preventing any data from '
                  f'being saved.', 'Please close it and try again.', sep='\n')

            return False



import unittest
import os
from src.app.image_processor import ImageProcessor


class TestSingleton(unittest.TestCase):
    def setUp(self) -> None:
        self.processor = ImageProcessor()

    def test_same_singleton(self):
        """
        Tests that multiple references point to the same instance of the singleton.
        """

        self.singleton: ImageProcessor = ImageProcessor()
        self.assertEqual(self.singleton, self.processor)

    def test_singleton_attributes(self):
        self.assertTrue(self.processor.urls == [])
        self.assertTrue(self.processor.path == os.path.join(os.path.expanduser('~'), 'Desktop', 'Ovative Group Caption '
                                                                                                'Generator'))



import sys

from src import utils
from src.app.image_processor import ImageProcessor


"""
This file is used to start the application. The start method is used in the runner.py file, which will ask the user for 
input once started.
"""


def start():
    processor: ImageProcessor = ImageProcessor()
    utils.clear()  # clear terminal for fresh start
    print('Please provide an image URL(s) or a directory path to an image(s). If submitting '
          'multiple URLs, separate them with a comma and a space (i.e, ",").\n(You may quit the application at '
          'anytime by entering "q" or "quit")')
    user_input: str = input('\n> ')  # collect user input

    inputs: list

    if len(user_input) != 0:
        inputs = [x.strip() for x in user_input.split(',')]  # split the input into a list using a comma as a delimiter
    else:
        # an empty input was given
        utils.clear()
        print('Please provide an image URL(s) or a directory to images. The application will terminate. Please try '
              'again.')
        sys.exit()

    # start looping over the given image sources (if applicable)
    while len(inputs) != 0:
        if utils.quitting(user_input):
            utils.end()
            return

        success: bool = processor.process_input(inputs)

        inputs.clear()  # clearing elements to reprompt users to possibly provide more images to keep the while loop
        reprompt(inputs, success)


def reprompt(inputs: list[str], success: bool) -> None:
    """
    Asks the user if they would like to continue using the application. If not, the application terminates.
    :param success: represents a successful completion of the image processing
    :param inputs: list of inputs from the previous cycle of images processed
    :return: None
    """

    msg: str = 'Would you like to provide more images? (y/n)' if success else 'Would you like to try again?'

    if len(inputs) == 0:
        print(f'\n{msg}')
        user_input = input('\n> ')

        if user_input.lower() in ['n', 'no', 'q', 'quit']:
            utils.end()

        utils.clear()
        print('Please provide an image URL(s) or a directory path to an image(s). If submitting '
              'multiple URLs, separate them with a comma and a space (i.e, ",").\n(You may quit the application at '
              'anytime by entering "q" or "quit")')

        user_input: str = input('\n> ')  # collect user input

        if utils.quitting(user_input):
            utils.end()

        inputs += [x.strip() for x in user_input.split(',')]


from PIL import Image


def caption_image(img: Image, processor, model) -> tuple[str, str]:
    """
    This method will caption the given Pillow Image using the given model, which is the Salesforce captioning model.
    Two captions will be generated and outputted.
    :param img: the image object used to be captioned
    :param processor: the processor used to caption the images
    :param model: the model used to caption the images
    :return: None
    """

    # Conditional image captioning
    text = "a photo of"
    inputs = processor(img, text, return_tensors="pt")

    out = model.generate(**inputs, max_length=60)
    con_caption: str = 'Conditional image caption: ' + processor.decode(out[0], skip_special_tokens=True)

    # Unconditional image captioning
    inputs = processor(images=img, return_tensors="pt")

    out = model.generate(**inputs, max_length=60)
    unc_caption: str = 'Unconditional image caption: ' + processor.decode(out[0], skip_special_tokens=True)

    return con_caption, unc_caption  # return both generated captions



from setuptools import setup, find_packages

VERSION = '1.0.1'
DESCRIPTION = 'An image caption generator for advertisements.'
LONG_DESCRIPTION = ('A package that generates captions, collects text, and saves it to a .csv file after being given '
                    'an image of an advertisement.')

# Setting up
setup(
    name='ogicg',
    version=VERSION,
    author='Ian King, Mason Myles, Omar Elsegeiny, Emily Anderson, & Bohdan Ivanyshyn',
    author_email="kingian404@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['pillow', 'torch', 'requests', 'validators', 'transformers', 'cohere',
                      'numpy', 'opencv-python-headless', 'gradio', 'easyocr'],
    keywords=['python', 'image', 'caption', 'generator', 'image caption generator', 'OCR',
              'optical character recognition', 'ad', 'ads', 'advertisement', 'advertisements',
              'ogicg', 'Ovative', 'Ovative Group', 'Ovative Group Image Caption Generator'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)
